/*

  libkohn_rtsp - Copyright 2011-2014 by Michael Kohn
  http://www.mikekohn.net/
  mike@mikekohn.net

  Source distributed under the LGPL license.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifndef WINDOWS
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <time.h>
#ifdef WINDOWS
#include <windows.h>
#include <winsock.h>
#endif


// http://www.ietf.org/rfc/rfc2326.txt 10.4

#include "kohn_rtsp.h"
#include "kohn_rtp_thread.h"

#define BUFFER_SIZE 1024

#define STARTS_WITH(a,b) (strncasecmp((char *)a,b,sizeof(b)-1)==0)

struct _read_context
{
  int in_ptr;
  int in_len;
  unsigned char in_buffer[BUFFER_SIZE];
  unsigned char out_buffer[BUFFER_SIZE];
};

//static struct _kohn_rtsp kohn_rtsp[KOHN_RTSP_MAX_CHANNELS] = {{ 0 }};

static void base64_encode(char *auth, char *user)
{
char base64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZ"
              "abcdefghijklmnopqrstuvwxyz"
              "0123456789+/";
int len;
int holding;
int ptr=0;
int i;

  len=strlen(user);

  for (i=0; i<len-(len%3); i=i+3)
  {
    holding=(user[i]<<16)|(user[i+1]<<8)|(user[i+2]);
    auth[ptr]=base64[holding>>18];
    auth[ptr+1]=base64[(holding>>12)&63];
    auth[ptr+2]=base64[(holding>>6)&63];
    auth[ptr+3]=base64[holding&63];
    ptr=ptr+4;
  }

  switch(len%3)
  {
    case 2:
      holding=(user[i]<<16)|(user[i+1]<<8);
      auth[ptr]=base64[holding>>18];
      auth[ptr+1]=base64[(holding>>12)&63];
      auth[ptr+2]=base64[(holding>>6)&63];
      auth[ptr+3]='=';
      break;
    case 1:
      holding=(user[i]<<16);
      auth[ptr]=base64[holding>>18];
      auth[ptr+1]=base64[(holding>>12)&63];
      auth[ptr+2]='=';
      auth[ptr+3]='=';
      break;
    default:
      ptr-=4;
      break;
  }

  auth[ptr+4]=0;
}

#if 0
static int parse_hex(char *s)
{
int n=0;

  while(1)
  {
    *s=tolower(*s);

    if (*s>='0' && *s<='9')
    {
      n=(n<<4)+((*s)-'0');
    }
      else
    if (*s>='a' && *s<='f')
    {
      n=(n<<4)+(((*s)-'a')+10);
    }
      else
    { break; }

    s++;
  }

  return n;
}
#endif

static void get_size(char *s, int *width, int *height)
{
char *a=s;
int w,h;

  while (*s!='x' && *s!=0) s++;
  if (*s==0) return;
  s++;

  w=atoi(a);
  h=atoi(s);

  if (w==0 || h==0) return;

  *width=w;
  *height=h;
}

static int check_status(unsigned char *s)
{
  if (!STARTS_WITH(s,"RTSP/")) return 0;

  while(*s!=0 && *s!=' ') s++; 
  if (*s==0) return 0;
  s++;

  return atoi((char *)s);
}

static int get_session(char *d, char *s)
{
int ptr=0;

  while(s[ptr]!=0)
  {
    if ((s[ptr]>='a' && s[ptr]<='f') ||
        (s[ptr]>='A' && s[ptr]<='F') ||
        (s[ptr]>='0' && s[ptr]<='9'))
    {
      d[ptr]=s[ptr];
      ptr++;
      if (ptr==16) return -1;
    }
      else
    {
      break;
    }
  }

  d[ptr]=0;

  return 0;
}

static int get_server_ports(char *s)
{
int port;

  s=strstr(s,"server_port=");
  if (s==NULL) return 0;

  s=s+sizeof("server_port=")-1;
  char *c=s;
  while (*c!='-' && *c!=0) c++;
  if (c==0) return 0;

  //*c=0;
  c++;

  port=atoi(s);
  if (port+1!=atoi(c)) return 0;

  return port;
}

static int get_server_info(char *s, char *name)
{
  s=strstr(s,name);
  if (s==NULL) return 0;

  s=s+strlen(name)+1;
  return atoi(s);
}

static int get_server_info_h(char *s, char *name)
{
int n;
  s=strstr(s,name);
  if (s==NULL) return 0;

  s=s+strlen(name)+1;
  n=strtol(s,NULL,16);
  //return parse_hex(s);
  return n;
}

static int open_two_udp_ports(struct _kohn_rtsp *kohn_rtsp)
{
int ports_tested=0;
int port=KOHN_RTSP_UDP_PORT_START;
struct sockaddr_in source_addr;

  kohn_rtsp->rtp_sockfd=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (kohn_rtsp->rtp_sockfd<0)
  {
    return KOHN_RTSP_SOCKET_ERROR;
  }
  kohn_rtsp->rtcp_sockfd=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (kohn_rtsp->rtcp_sockfd<0)
  {
    close(kohn_rtsp->rtp_sockfd);
    return KOHN_RTSP_SOCKET_ERROR;
  }

  // FIXME - can this thing be reused like this?
  memset(&source_addr, 0, sizeof(source_addr));
  source_addr.sin_family=AF_INET;
  source_addr.sin_addr.s_addr=htonl(INADDR_ANY);

  while(ports_tested<KOHN_RTSP_MAX_UDP_PORT_TEST)
  {
    kohn_rtsp->client_ports=port;
    source_addr.sin_port=htons(port++);

    if (bind(kohn_rtsp->rtp_sockfd, (struct sockaddr *)&source_addr, sizeof(source_addr))==-1)
    {
#ifdef DEBUG
    printf("RTP bind fail for port %d\n", port-1);
#endif
      port++;
      ports_tested+=2;
      continue;
    }

#ifdef DEBUG
    printf("RTP bound on port %d\n", port-1);
#endif

    source_addr.sin_port=htons(port++);
    if (bind(kohn_rtsp->rtcp_sockfd, (struct sockaddr *)&source_addr, sizeof(source_addr))==-1)
    {
#ifdef DEBUG
    printf("RTCP bind fail for port %d\n", port-1);
#endif
      ports_tested+=2;
      continue;
    }

#ifdef DEBUG
    printf("RTCP bound on port %d\n", port-1);
#endif
     break;
  }

  if (ports_tested==KOHN_RTSP_MAX_UDP_PORT_TEST) return KOHN_RTSP_SOCKET_ERROR;

  /* Some retarded OS's (aka Windows XP and such) make this thing small */
  int len;
  socklen_t socklen=sizeof(len);
  getsockopt(kohn_rtsp->rtp_sockfd, SOL_SOCKET, SO_RCVBUF, &len, &socklen);

#ifdef DEBUG
  printf(":: UDP buffer size=%d\n", len);
#endif

  if (len<65536)
  {
    len = 65536;   /* 64k UDP buffer size.  Should this be bigger? */
    setsockopt(kohn_rtsp->rtp_sockfd, SOL_SOCKET, SO_RCVBUF, &len, sizeof(len));
  }


  return KOHN_RTSP_OKAY;
}

static int send_data(int sockfd, char *message, int message_len)
{
//struct timeval tv;
fd_set writeset;
int t;

  FD_ZERO(&writeset);
  FD_SET(sockfd,&writeset);

  // FIXME - Time out of null is bad bad bad

  //tv.tv_sec=1;
  //tv.tv_usec=0;

  if (((t=select(sockfd+1,NULL,&writeset,NULL,NULL))==-1) && (errno!=EINTR))
  {
    printf("Problem with select\n");
    return -1;
  }
    else
  {
    if (t==0) return -1;
    t=send(sockfd,message,message_len,0);
#ifdef WINDOWS
    _commit(sockfd);
#else
    fsync(sockfd);
#endif
    return t;
  }

  return 0;
}

static int buffered_read(int sockfd, struct _read_context *rc)
{
int r,out_ptr=0;
int bytes_consumed=0;

  while(1)
  {
    errno=0;
    if (rc->in_ptr>=rc->in_len)
    {
      rc->in_len=recv(sockfd,rc->in_buffer,BUFFER_SIZE,0);
      rc->in_ptr=0;
      if (rc->in_len==0) return -1;
      if (rc->in_len<0)
      {
        rc->in_len=0;
        return -1;
      }
    }

    for (r=rc->in_ptr; r<rc->in_len; r++)
    {
      bytes_consumed++;
      if (rc->in_buffer[r]=='\r') continue;
      if (rc->in_buffer[r]=='\n')
      {
        rc->out_buffer[out_ptr]=0;
        rc->in_ptr=r+1;
        return bytes_consumed;
      }

      if (rc->in_buffer[r]==127)
      { if (out_ptr>0) out_ptr--; }
        else
      { if (rc->in_buffer[r]>=' ' && rc->in_buffer[r]<254) rc->out_buffer[out_ptr++]=rc->in_buffer[r]; }

      if (out_ptr==BUFFER_SIZE-1)
      {
        rc->out_buffer[out_ptr]=0;
        rc->in_ptr=r+1;
        return bytes_consumed;
      }
    }
  }
}

static int kohn_rtsp_parse_url(char *url, struct _kohn_rtsp_url *kohn_rtsp_url)
{
char *private_data=kohn_rtsp_url->private_data;
int n;

  memset(kohn_rtsp_url, 0, kohn_rtsp_url->private_data-(char *)kohn_rtsp_url);
  kohn_rtsp_url->port=554;

  // Is URL too long for us?
  if (strlen(url)+2>1024) return -2;

  if (strncasecmp(url, "rtsp://", 7)!=0)
  {
    // Not rtsp stream?
    return -1;
  }

  url=url+7;

  // Get filename to request
  n=0;
  while(url[n]!=0 && url[n]!='/') n++;
  if (url[n]==0)
  {
    private_data[0]='/';
    private_data[1]=0;
    private_data=private_data+2;
    strcpy(private_data, url);
  }
    else
  {
    strcpy(kohn_rtsp_url->private_data, url+n);
    private_data=private_data+strlen(url+n)+1;
    strncpy(private_data, url, n);
    private_data[n]=0;
  }
  kohn_rtsp_url->file=kohn_rtsp_url->private_data;

  // Check for username:password
  n=0;
  while(private_data[n]!=0 && private_data[n]!='@') n++;
  if (private_data[n]=='@')
  {
    private_data[n]=0;
    kohn_rtsp_url->user=private_data;
    private_data=private_data+n+1;
  }

  kohn_rtsp_url->host=private_data;

  // Check for port
  n=0;
  while(private_data[n]!=0 && private_data[n]!=':') n++;
  if (private_data[n]==':')
  {
    private_data[n]=0;
    kohn_rtsp_url->port=atoi(private_data+n+1);
  }

  return 0;
}

int kohn_rtsp_open(struct _kohn_rtsp **kohn_rtsp_ret, char *url, kohn_rtsp_callback_t data_callback, kohn_rtsp_callback_t rtcp_callback, int mode, void *context)
{
//struct _kohn_rtsp_url kohn_rtsp_url;
struct _kohn_rtsp *kohn_rtsp=NULL;
struct sockaddr_in cli_addr;
struct _read_context rc = { 0 };
//int channel;
int sockfd;
int status_code;

/*
  // Find free channel
  for (channel=0; channel<KOHN_RTSP_MAX_CHANNELS; channel++)
  {
    if (kohn_rtsp[channel].inuse==0)
    {
      *channel_out=channel;
      break;
    }
  }

  if (channel==KOHN_RTSP_MAX_CHANNELS)
  {
    return KOHN_RTSP_FULL;
  }
*/

  kohn_rtsp=(struct _kohn_rtsp *)malloc(sizeof(struct _kohn_rtsp));
  *kohn_rtsp_ret=kohn_rtsp;

  if (kohn_rtsp==NULL) { return KOHN_RTSP_GENERAL_ERROR; }

  memset(kohn_rtsp, 0, sizeof(struct _kohn_rtsp));
  kohn_rtsp->context=context;
  kohn_rtsp->mode=mode;
  kohn_rtsp->width=640;
  kohn_rtsp->height=480;
  kohn_rtsp->fps=30.0;
  strcpy(kohn_rtsp->fourcc,"xvid");
  if (kohn_rtsp_parse_url(url, &kohn_rtsp->url)!=0)
  {
    return KOHN_RTSP_BAD_URL;
  }
  kohn_rtsp->ssrc_local=time(NULL);

#ifdef DEBUG
  printf("   file: '%s'\n", kohn_rtsp->url.file);
  printf("   user: '%s'\n", kohn_rtsp->url.user==NULL?"":kohn_rtsp->url.user);
  printf("   host: '%s'\n", kohn_rtsp->url.host);
  printf("   port: %d\n", kohn_rtsp->url.port);
  //printf("channel: %d\n", channel);
#endif

  char extra_headers[256];
  if (kohn_rtsp->url.user!=NULL)
  {
    char auth[128];
    base64_encode(auth, kohn_rtsp->url.user);
    sprintf(extra_headers,"Authorization: Basic %s\r\n", auth);
  }
    else
  { extra_headers[0]=0; }

  kohn_rtsp->rtp_callback=data_callback;
  kohn_rtsp->rtcp_callback=rtcp_callback;

  if ((sockfd=socket(AF_INET,SOCK_STREAM,0))<0)
  {
    return KOHN_RTSP_SOCKET_ERROR;
  }

  memset(&cli_addr,0,sizeof(cli_addr));
  cli_addr.sin_family=AF_INET;
  cli_addr.sin_port=htons(kohn_rtsp->url.port);
  if (inet_aton(kohn_rtsp->url.host, &cli_addr.sin_addr)==0)
  {
    return KOHN_RTSP_UNKNOWN_HOST;
  }

  memcpy(&kohn_rtsp->sin_addr, &cli_addr.sin_addr, sizeof(struct in_addr));

  if (connect(sockfd,(struct sockaddr *)&cli_addr,sizeof(cli_addr))<0)
  {
    return KOHN_RTSP_SOCKET_ERROR;
  }

  char message[2048];
  //sprintf(message, "OPTIONS * RTSP/1.0\nRequire: method.announce\nCSeq: 1\n\n");
  sprintf(message, "OPTIONS * RTSP/1.0\r\nCSeq: 1\r\n\r\n");

#ifdef DEBUG
printf("%s",message);
#endif

  send_data(sockfd, message, strlen(message));

  rc.in_ptr=0;
  int blank_count=0;
  status_code=0;
  while(1)
  {
    if (buffered_read(sockfd, &rc)==-1) return KOHN_RTSP_SOCKET_ERROR;
#ifdef DEBUG
    printf("in> '%s'\n", rc.out_buffer);
#endif
    if (status_code==0)
    {
      status_code=check_status(rc.out_buffer);
      if (status_code==404) { return KOHN_RTSP_404; }
      if (status_code!=200) { return KOHN_RTSP_GENERAL_ERROR; }
    }

    if (rc.out_buffer[0]==0) blank_count++;
    if (blank_count==1) break;
  }

  sprintf(message, "DESCRIBE rtsp://%s:%d%s RTSP/1.0\r\nAccept: application/sdp\r\nCSeq: 2\r\n%s\r\n", kohn_rtsp->url.host, kohn_rtsp->url.port, kohn_rtsp->url.file,extra_headers);

#ifdef DEBUG
  printf("%s\n", message);
#endif
  send_data(sockfd, message, strlen(message));

  blank_count=0;
  int bytes_consumed=0;
  int content_length=0;
  status_code=0;
  while(1)
  {
    int l=buffered_read(sockfd, &rc);
    if (l<0) return KOHN_RTSP_SOCKET_ERROR;
#ifdef DEBUG
    printf("in> '%s'\n", rc.out_buffer);
#endif
    if (status_code==0)
    {
      status_code=check_status(rc.out_buffer);
      if (status_code==404) { return KOHN_RTSP_404; }
      if (status_code!=200) { return KOHN_RTSP_GENERAL_ERROR; }
    }
      else
    if (blank_count==0)
    {
      if (STARTS_WITH(rc.out_buffer, "content-length:"))
      {
        content_length=atoi((char *)rc.out_buffer+sizeof("content-length:"));
      }
    }
      else
    if (blank_count==1)
    {
      if (STARTS_WITH(rc.out_buffer, "a=framerate:"))
      {
        kohn_rtsp->fps=atof((char *)rc.out_buffer+sizeof("a=framerate:")-1);
      }
        else
      if (STARTS_WITH(rc.out_buffer, "a=rtpmap:"))
      {
        if (strstr((char *)rc.out_buffer, "H264")!=NULL ||
            strstr((char *)rc.out_buffer, "h264")!=NULL)
        {
          strcpy(kohn_rtsp->fourcc, "h264");
        }
      }
        else
      if (strstr((char *)rc.out_buffer, "resolution=")!=NULL)
      {
        get_size(strstr((char *)rc.out_buffer, "resolution=")+sizeof("resolution=")-1, &kohn_rtsp->width, &kohn_rtsp->height);
      }

      bytes_consumed+=l;
      if (bytes_consumed>=content_length) break;
    }
    if (rc.out_buffer[0]==0) blank_count++;
    if (blank_count==2) break;
  }

  int ret=open_two_udp_ports(kohn_rtsp);
  if (ret!=KOHN_RTSP_OKAY) return ret;

  // RTSP SETUP
  sprintf(message, "SETUP rtsp://%s:%d%s RTSP/1.0\r\nCSeq: 3\r\nTransport: RTP/AVP;unicast;client_port=%d-%d\r\n%s\r\n", kohn_rtsp->url.host, kohn_rtsp->url.port, kohn_rtsp->url.file,kohn_rtsp->client_ports,kohn_rtsp->client_ports+1,extra_headers);

#ifdef DEBUG
  printf("%s\n", message);
#endif

  send_data(sockfd, message, strlen(message));

  //rc.in_ptr=0;
  blank_count=0;
  status_code=0;
  while(1)
  {
    if (buffered_read(sockfd, &rc)==-1) return KOHN_RTSP_SOCKET_ERROR;
#ifdef DEBUG
    printf("in> '%s'\n", rc.out_buffer);
#endif
    if (status_code==0)
    {
      status_code=check_status(rc.out_buffer);
      if (status_code==404) { return KOHN_RTSP_404; }
      if (status_code!=200) { return KOHN_RTSP_GENERAL_ERROR; }
    }
      else
    if (STARTS_WITH(rc.out_buffer,"Session:"))
    {
      if (get_session(kohn_rtsp->session, (char *)rc.out_buffer+sizeof("Session:"))!=0) return KOHN_RTSP_GENERAL_ERROR;
    }
    if (STARTS_WITH(rc.out_buffer,"Transport:"))
    {
      kohn_rtsp->server_ports=get_server_ports((char *)rc.out_buffer);
      kohn_rtsp->ssrc=get_server_info_h((char *)rc.out_buffer, "ssrc");
      if (kohn_rtsp->server_ports==0) return KOHN_RTSP_BAD_PORTS;
    }
    if (rc.out_buffer[0]==0) blank_count++;
    if (blank_count==1) break;
  }

#ifdef DEBUG
  printf("Client ports: %d-%d\n", kohn_rtsp->client_ports, kohn_rtsp->client_ports+1);
  printf("Server ports: %d-%d\n", kohn_rtsp->server_ports, kohn_rtsp->server_ports+1);
  printf("        SSRC: %d\n", kohn_rtsp->ssrc);
#endif

  // Make sure we are ready to catch packets before it starts.
  kohn_rtsp_start_rtp(kohn_rtsp);

  sprintf(message, "PLAY rtsp://%s:%d%s RTSP/1.0\r\nCSeq: 4\r\nSession: %s\r\n%s\r\n", kohn_rtsp->url.host, kohn_rtsp->url.port, kohn_rtsp->url.file,kohn_rtsp->session,extra_headers);

#ifdef DEBUG
  printf("%s\n", message);
#endif

  send_data(sockfd, message, strlen(message));

  //rc.in_ptr=0;
  blank_count=0;
  status_code=0;
  while(1)
  {
    if (buffered_read(sockfd, &rc)==-1) return KOHN_RTSP_SOCKET_ERROR;
#ifdef DEBUG
    printf("in> '%s'\n", rc.out_buffer);
#endif
    if (rc.out_buffer[0]==0) blank_count++;
    if (status_code==0)
    {
      status_code=check_status(rc.out_buffer);
      if (status_code==404) { return KOHN_RTSP_404; }
      if (status_code!=200) { return KOHN_RTSP_GENERAL_ERROR; }
    }
      else
    if (STARTS_WITH(rc.out_buffer,"RTP-Info:"))
    {
      kohn_rtsp->seq=get_server_info((char *)rc.out_buffer, "seq");
    }
    if (blank_count==1) break;
  }

#ifdef DEBUG
  printf("Sequence Num: %d\n", kohn_rtsp->seq);
#endif

  //kohn_rtsp->channel=channel; // FIXME - is there a better way?
  kohn_rtsp->inuse=1;

  close(sockfd);

  return KOHN_RTSP_OKAY;
}

int kohn_rtsp_is_running(struct _kohn_rtsp *kohn_rtsp)
{
  return kohn_rtsp->inuse;
}

int kohn_rtsp_stop(struct _kohn_rtsp *kohn_rtsp)
{
struct sockaddr_in cli_addr;
struct _read_context rc={ 0 };
int sockfd;
int status_code;

  if (kohn_rtsp->inuse==0) { return KOHN_RTSP_GENERAL_ERROR; }

//#ifdef DEBUG
//  printf("rc: in_ptr=%d in_len=%d.\n", rc.in_ptr, rc.in_len);
//#endif

  kohn_rtsp->inuse=3;

  if ((sockfd=socket(AF_INET,SOCK_STREAM,0))<0)
  {
    return KOHN_RTSP_SOCKET_ERROR;
  }

  memset(&cli_addr,0,sizeof(cli_addr));
  cli_addr.sin_family=AF_INET;
  cli_addr.sin_port=htons(kohn_rtsp->url.port);

  if (inet_aton(kohn_rtsp->url.host, &cli_addr.sin_addr)==0)
  {
    return KOHN_RTSP_UNKNOWN_HOST;
  }

  if (connect(sockfd,(struct sockaddr *)&cli_addr,sizeof(cli_addr))<0)
  {
    return KOHN_RTSP_SOCKET_ERROR;
  }

  char extra_headers[256];
  if (kohn_rtsp->url.user!=NULL)
  {
    char auth[128];
    base64_encode(auth, kohn_rtsp->url.user);
    sprintf(extra_headers,"Authorization: Basic %s\n", auth);
  }
    else
  { extra_headers[0]=0; }

  char message[2048];
  sprintf(message, "TEARDOWN rtsp://%s:%d%s RTSP/1.0\nCSeq: 5\nSession: %s\n%s\n", kohn_rtsp->url.host, kohn_rtsp->url.port, kohn_rtsp->url.file,kohn_rtsp->session,extra_headers);

#ifdef DEBUG
  printf("%s\n", message);
#endif

  send_data(sockfd, message, strlen(message));

  //rc.in_ptr=0;
  int blank_count=0;
  status_code=0;
  while(1)
  {
    if (buffered_read(sockfd, &rc)==-1) return KOHN_RTSP_SOCKET_ERROR;
#ifdef DEBUG
    printf("in> '%s'\n", rc.out_buffer);
#endif
    if (status_code==0)
    {
      status_code=check_status(rc.out_buffer);
      if (status_code==404) { return KOHN_RTSP_404; }
      if (status_code!=200) { return KOHN_RTSP_GENERAL_ERROR; }
    }
    if (rc.out_buffer[0]==0) blank_count++;
    if (blank_count==1) break;
  }

  return KOHN_RTSP_OKAY;
}

int kohn_rtsp_close(struct _kohn_rtsp *kohn_rtsp)
{
  kohn_rtsp_stop(kohn_rtsp);

  free(kohn_rtsp);
  return KOHN_RTSP_OKAY;
}

float kohn_rtsp_get_framerate(struct _kohn_rtsp *kohn_rtsp)
{
  return kohn_rtsp->fps;
}

const char *kohn_rtsp_get_codec(struct _kohn_rtsp *kohn_rtsp)
{
  return kohn_rtsp->fourcc;
}

int kohn_rtsp_get_width(struct _kohn_rtsp *kohn_rtsp)
{
  return kohn_rtsp->width;
}

int kohn_rtsp_get_height(struct _kohn_rtsp *kohn_rtsp)
{
  return kohn_rtsp->height;
}

int kohn_rtsp_get_packet_count(struct _kohn_rtsp *kohn_rtsp)
{
  return kohn_rtsp->packet_count;
}



