/*

  libkohn_rtsp - Copyright 2011-2014 by Michael Kohn
  http://www.mikekohn.net/
  mike@mikekohn.net

  Source distributed under the LGPL license.

*/

#ifndef KOHN_RTSP_H
#define KOHN_RTSP_H

#ifndef WINDOWS
#include <netinet/in.h>
#endif

// All data passed to callback removes RTP headers.
#define KOHN_RTP_MODE_RAW 0     // Pass all rtp packets callback as they come
#define KOHN_RTP_MODE_RTP 2     // Pass entire rtp packet as is
#define KOHN_RTP_MODE_ORDER 1   // Pass all rtp packets callback in order
#define KOHN_RTP_MODE_MARKER 3  // Pass all packets assembled up to markers

#define KOHN_RTSP_OKAY 0
#define KOHN_RTSP_404 -1
#define KOHN_RTSP_FULL -2
#define KOHN_RTSP_SOCKET_ERROR -3
#define KOHN_RTSP_UNKNOWN_HOST -4
#define KOHN_RTSP_BAD_PORTS -5
#define KOHN_RTSP_GENERAL_ERROR -6
#define KOHN_RTSP_BAD_URL -7

#define KOHN_RTSP_MAX_CHANNELS 20
#define KOHN_RTSP_MAX_UDP_PORT_TEST 5000
#define KOHN_RTSP_UDP_PORT_START 8000

#define KOHN_RTSP_FRAME_DATA_START_LEN 32768
#define KOHN_RTSP_FRAME_DATA_GROW_LEN 32768
#define KOHN_RTSP_FRAME_DATA_MAX_LEN 131072

typedef void (*kohn_rtsp_callback_t)(unsigned char *data, int len, void *context);

struct _kohn_rtsp_url
{
  char *user;
  char *host;
  char *file;
  int port;
  char private_data[1024];
};

struct _kohn_rtsp
{
  void *context;
  int inuse;
  char fourcc[8];        // why 8?  might as well align since 5 is min
  int width;
  int height;
  float fps;
  char mode;             // KOHN_RTSP_MODE_*
  char session[16];
  int client_ports;      // ports are client_ports and client_ports+1
  int server_ports;      // ports are server_ports and server_ports+1
  struct in_addr sin_addr;
  int rtp_sockfd;
  int rtcp_sockfd;
  int fraction_lost;
  int packet_count;
  unsigned int ssrc;
  unsigned int ssrc_local;
  unsigned short int seq;
  unsigned short int extended_seq;
  unsigned int last_sr;
  time_t rtcp_sr_time;
  time_t rtcp_rr_time;
  int rtcp_sr_count;
  kohn_rtsp_callback_t rtp_callback;
  kohn_rtsp_callback_t rtcp_callback;
  struct _kohn_rtsp *next;
  struct _kohn_rtsp_url url;
  int frame_data_len;
  int frame_data_ptr;
  unsigned char *frame_data;
  struct _buffered_rtp *rtp_head;
  int buffered_count;
};

// void kohn_rtsp_parse_url(char *url, struct _kohn_rtsp_url *kohn_rtsp_url);
//int kohn_rtsp_allocate(int channels);

int kohn_rtsp_open(struct _kohn_rtsp **kohn_rtsp, char *url, kohn_rtsp_callback_t data_callback, kohn_rtsp_callback_t rtcp_callback, int mode, void *context);
int kohn_rtsp_stop(struct _kohn_rtsp *kohn_rtsp);
int kohn_rtsp_close(struct _kohn_rtsp *kohn_rtsp);
int kohn_rtsp_is_running(struct _kohn_rtsp *kohn_rtsp);
float kohn_rtsp_get_framerate(struct _kohn_rtsp *kohn_rtsp);
const char *kohn_rtsp_get_codec(struct _kohn_rtsp *kohn_rtsp);
int kohn_rtsp_get_width(struct _kohn_rtsp *kohn_rtsp);
int kohn_rtsp_get_height(struct _kohn_rtsp *kohn_rtsp);
int kohn_rtsp_get_packet_count(struct _kohn_rtsp *kohn_rtsp);

#endif


