/*

  libkohn_rtsp - Copyright 2011-2014 by Michael Kohn
  http://www.mikekohn.net/
  mike@mikekohn.net

  Source distributed under the LGPL license.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

#ifdef KOHN_AVI
#include <kohn_avi.h>
#endif

#include "kohn_rtsp.h"

struct _rtsp_rec
{
#ifdef KOHN_AVI
  struct kavi_t *kavi;
#endif
  FILE *out;
  int frame_count;
};

struct _rtsp_rec rtsp_rec[20];
static int running=1;

void handler(int signal_num)
{
  printf("Signal caught\n");
  running=0;
}

void callback(unsigned char *data, int data_len, void *context)
{
int n;
int id=*((int *)context);
#ifdef DEBUG
  printf("--- Data on channel %d of len %d\n", id, data_len);
#endif

#ifdef SUPER_DEBUG
  for (n=0; n<data_len; n++)
  {
    if ((n%16)==0) printf("\n");
    printf("%02x ", data[n]);
  }
  printf("\n");
#endif

  n=fwrite(data, 1, data_len, rtsp_rec[id].out);
  if (n==0) { return; }  // FIXME - I hate this bullshit :(
#ifdef KOHN_AVI
  kavi_add_frame(rtsp_rec[id].kavi, data, data_len);
#endif
  rtsp_rec[id].frame_count++;
}

int main(int argc, char *argv[])
{
struct _kohn_rtsp *kohn_rtsp;
int channel=0;

  if (argc!=2)
  {
    printf("test_rtsp - Copyright 2011-2014 by Michael Kohn\n");
    printf("Version: 2014-01-08\n");
    printf("Usage: %s <URL>\n", argv[0]);
    exit(0);
  }

  char filename[256];

  sprintf(filename,"out_%d.mpeg",channel);

  rtsp_rec[channel].frame_count=0;
  rtsp_rec[channel].out=fopen(filename, "wb");
  if (rtsp_rec[channel].out==NULL)
  {
    printf("Couldn't open %s.\n",filename);
    exit(1);
  }

#ifdef KOHN_AVI
  sprintf(filename,"out_%d.avi",channel);
  rtsp_rec[channel].kavi=kavi_open(filename, 640, 480, "xvid", 30, NULL);

  if (rtsp_rec[channel].kavi==NULL)
  {
    printf("Couldn't open %s.\n",filename);
    exit(1);
  }
#endif

  //int status=kohn_rtsp_open(&kohn_rtsp, argv[1], callback, NULL, KOHN_RTP_MODE_RAW, channel);
  int status=kohn_rtsp_open(&kohn_rtsp, argv[1], callback, NULL, KOHN_RTP_MODE_MARKER, &channel);
  if (status!=0)
  {
    printf("Error %d opening stream\n", status);
    exit(1);
  }

  signal(SIGTERM,handler);
  signal(SIGINT,handler);

#ifdef DEBUG
  printf("Opened rtsp stream on channel %d\n", channel);
#endif

  int count=20;
  while(count>0 && running==1)
  {
    count--;
    sleep(1);
  }

  printf("Stopping\n");
  kohn_rtsp_stop(kohn_rtsp);
  printf("Stopped\n");

#ifdef KOHN_AVI
  // Correct the framerate and codec from the SDP info
  kavi_set_framerate(rtsp_rec[channel].kavi, kohn_rtsp_get_framerate(kohn_rtsp));
  kavi_set_codec(rtsp_rec[channel].kavi, kohn_rtsp_get_codec(kohn_rtsp));
  kavi_set_size(rtsp_rec[channel].kavi, kohn_rtsp_get_width(kohn_rtsp), kohn_rtsp_get_height(kohn_rtsp));
#endif

#ifdef DEBUG
  printf("     Size: %dx%d\n", kohn_rtsp_get_width(kohn_rtsp), kohn_rtsp_get_height(kohn_rtsp));
  printf("Framerate: %f\n", kohn_rtsp_get_framerate(kohn_rtsp));
  printf("    Codec: %s\n", kohn_rtsp_get_codec(kohn_rtsp));
#endif

  printf("Recorded %d frames from %d rtp packets.\n", rtsp_rec[channel].frame_count, kohn_rtsp_get_packet_count(kohn_rtsp));

  fclose(rtsp_rec[channel].out);
#ifdef KOHN_AVI
  kavi_close(rtsp_rec[channel].kavi);
#endif

  int n=5;
  while(n>=0)
  {
    if (kohn_rtsp_is_running(kohn_rtsp)==0) break;
    sleep(1);
  }

  kohn_rtsp_close(kohn_rtsp);

  return 0;
}



